;DUCK
;ATmega8 @ 8 MHz (internal R/C osc.)
;(C) Arne Rossius, http://www.ebps.de.vu/
;
;FUSES
;  low  = 0xE4 (8 MHz int. R/C, BOD disabled, slow start-up)
;  high = 0x99 (WDT always enabled)
;
;PORTS
;  PB0   = Push-button (to ground)
;  PB1-2 = Speaker PWM signals [OC1A, OC1B] (only PB1 used on prototype)
;  PB3   = Always-on input (active low, jumper on ISP header) [ISP]
;  PB4-5 = [unused, ISP]
;  PB6-7 = [unused]
;  PC0-5 = LED matrix columns 0~5 (active high)
;  PD0-5 = LED matrix rows 0~5 (active high, external NPN transistors)
;  PD6   = [unused]
;  PD7   = Audio amplifier enable (active low, protype only)
;
;VERSIONS
;  2015-12-15  Arne Rossius
;   * first version

.include "../include/m8def.inc"

.def	sreg_backup = R2
.def	repeat = R3
.def	leds = R4 ;used in mux interrupt
.def	spl_backup = R5
.def	sph_backup = R6

.def	temp = R16
.def	temp2 = R17
.def	count = R18
.def	inttmp = R19 ;used in interrupts
.def	intcnt = R20 ;used in interrupts
.def	pwm = R21 ;used in mux interrupt
.def	mux = R22 ;used in mux interrupt
.def	button = R23 ;used in mux interrupt
.def	wakeup_count = R24
.def	flags = R25
	.equ	fButtonShort = 0
	.equ	fButtonLong = 1
	.equ	fExpectWdr = 2
;R27:R26 used for X pointer
;R29:R28 used for Y pointer
;R31:R30 used for Z pointer

.equ	BUTTON_SHORT_TIME = 2
.equ	BUTTON_LONG_TIME = 80

;===============================================================================

.dseg
	RAM_Frame1: .byte 34 ;60
	RAM_Frame2: .byte 34 ;60

;===============================================================================

.cseg
.org 0x000
	rjmp	reset
	
.org OVF1addr
	;Timer 1 overflow (update PWM values for audio output): 31.25 kHz
	rjmp	ovf1

.org OVF0addr
	;Timer 0 overflow (LED PWM/Mux): interval according to current PWM step
	in	sreg_backup, SREG
	push	ZL
	push	ZH
	
	subi	pwm, 1
	brcc	ovf0_pwm
	
	;LED Multiplexing
	lsl	mux
	sbrs	mux, 6
	rjmp	mux_end
	ldi	mux, 0x01
	
	;reset watchdog (~82 Hz)
	wdr
	
	;read button (~82 Hz)
	sbic	PINB, 0
	rjmp	button_released
	inc	button ;button pressed
	cpi	button, BUTTON_LONG_TIME
	brlo	mux_end
	ldi	button, BUTTON_LONG_TIME
	brne	mux_end
	ori	flags, 1<<fButtonLong
	rjmp	mux_end
button_released:
	mov	inttmp, button
	clr	button
	cpi	inttmp, BUTTON_SHORT_TIME
	brlo	mux_end
	cpi	inttmp, BUTTON_LONG_TIME
	brsh	mux_end
	ori	flags, 1<<fButtonShort
mux_end:

	;LED PWM
	ldi	pwm, 6
	ldi	inttmp, -128 ;for PWM
	rjmp	ovf0_set_timer
ovf0_pwm:
	ldi	inttmp, -2
	sbrc	pwm, 2
	ldi	inttmp, -32
	sbrc	pwm, 1
	ldi	inttmp, -8
	sbrc	pwm, 0
	lsl	inttmp ;(multiply by 2)
	
ovf0_set_timer:
	;Set timer for next interrupt
	;(delay until next interrupt depends on PWM step for gamma correction)
	out	TCNT0, inttmp
	
	;set LEDs
	ldi	ZH, HIGH(RAM_Frame1-6)
	ldi	ZL, LOW(RAM_Frame1-6)
	mov	inttmp, mux
ovf0_addr_loop:
	adiw	ZH:ZL, 6
	lsr	inttmp
	brcc	ovf0_addr_loop
	clr	leds
	ldi	intcnt, 0x01
ovf0_pwm_loop:
	ld	inttmp, Z+
	cp	pwm, inttmp
	brsh	PC+2
	or	leds, intcnt
	lsl	intcnt
	sbrs	intcnt, 6
	rjmp	ovf0_pwm_loop
	
	ldi	inttmp, 0x00
	out	PORTC, inttmp ;LEDs off
	ldi	inttmp, 0x80
	or	inttmp, mux
	out	PORTD, inttmp ;switch row
	rjmp PC+1
	rjmp PC+1
	rjmp PC+1
	rjmp PC+1
	rjmp PC+1
	out	PORTC, leds ;LEDs on
	
	;return
	pop	ZH
	pop	ZL
	out	SREG, sreg_backup
	reti
	
;--------------------

ovf1:
	;timer 1 overflow interrupt: load PWM with next audio sample
	in	sreg_backup, SREG
	
	;read byte from flash and output to PWM registers
	lpm	inttmp, Z+
	cpi	inttmp, 0 ;no more data?
	breq	ovf1_stop
	out	OCR1AL, inttmp ;set new PWM value
	neg	inttmp
	out	OCR1BL, inttmp
	
ovf1_end:
	;return
	out	SREG, sreg_backup
	reti
	
ovf1_stop:
	;playback finished
	in	inttmp, TIMSK
	andi	inttmp, ~(1<<TOIE1)
	out	TIMSK, inttmp
	rjmp	ovf1_end
	
;===============================================================================

;animation instruction macros

.macro FILL ;FILL led, n, brt
  .if ((@0 < 1) || (@0 > 34))
    .error "FILL: LED number out of range (1~34)."
  .endif
  .if ((@1 < 1) || (@1 > 34))
    .error "FILL: stretch length out of range (1~34)."
  .endif
  .if ((@2 < 0) || (@2 > 7))
    .error "FILL: Brightness out of range (0~7)."
  .endif
  .db ((@0 - 1) << 1) | (@1 >> 5), LOW(@1 << 3) | @2
.endmacro

.macro LED ;LED led, brt
  FILL @0, 1, @1
.endmacro

.macro CLEAR ;CLEAR
  FILL 1, 34, 0
.endmacro

.macro ROTATE ;ROTATE n
  .if ((@0 >= 1) && (@0 <= 33))
    .db 0x80, @0
  .else
    .if ((@0 >= -33) && (@0 <= -1))
      .db 0x80, 34 + @0
    .else
      .error "ROTATE: value out of range (-33~-1 or 1~33)."
    .endif
  .endif
.endmacro

.macro FADE ;FADE n
  .if (((@0 >= 1) && (@0 <= 7)) || ((@0 >= -7) && ( @0 <= -1)))
    .db 0x81, @0
  .else
    .error "FADE: value out of range (-7~-7 or 1~7)."
  .endif
.endmacro

.macro DELAY ;DELAY n
  .if ((@0 < 10) || (@0 > 256 * 10))
    .error "DELAY: value out of range (10~2560)."
  .endif
  .db 0x82, LOW((@0 + 5) / 10)
.endmacro

.macro REPEAT ;REPEAT n
  .if ((@0 < 1) || (@0 > 255))
    .error "REPEAT: value out of range (1~255)."
  .endif
  .db 0x83, @0
.endmacro

.macro REPEAT_END ;REPEAT_END
  .db 0x83, 0
.endmacro

.macro ANI_END ;ANI_END
  .db 0x84, 0
.endmacro

;===============================================================================

animations:
	.include "animations/ani_list.asm"

audio:
	.include "audio/quack32500_u8.asm"
	
;===============================================================================

reset:
	;disable analog comparator
	ldi	temp, 0x80
	out	ACSR, temp
	
	;init ports
	ldi	temp, 0x09 ;enable pull-up for push button and always-on inputs
	out	PORTB, temp
	ldi	temp, 0xF6
	out	DDRB, temp
	ldi	temp, 0x00
	out	PORTC, temp
	ldi	temp, 0x3F
	out	DDRC, temp
	ldi	temp, 0x80 ;prototype: amplifier off
	out	PORTD, temp
	ldi	temp, 0xFF
	out	DDRD, temp

	;reset watchdog
	wdr
	
	;init watchdog timer (already enabled by fuse bit)
	ldi	temp, 1<<WDCE | 1<<WDE ;(timed sequence for access to WDTCR)
	out	WDTCR, temp
	ldi	temp, 1<<WDP2 ;256k cycles (~0.26 s @ 5 V)
	out	WDTCR, temp
	
	;check for watchdog reset
	in	temp, MCUCSR
	ldi	temp2, 0
	out	MCUCSR, temp
	sbrs	temp, WDRF
	rjmp	wdr_end
	sbrc	flags, fExpectWdr
	rjmp	wdr_wakeup
wdr_end:
	;init stack pointer
	ldi	temp, HIGH(RAMEND)
	out	SPH, temp
	ldi	temp, LOW(RAMEND)
	out	SPL, temp
	
	;init RAM
	ldi	ZH, HIGH(RAM_Frame1)
	ldi	ZL, LOW(RAM_Frame1)
	ldi	temp, 0
	ldi	count, 36 ;virtual matrix size is 6x6 LEDs
init_ram:
	st	Z+, temp
	dec	count
	brne	init_ram
	
	;init registers
	ldi	pwm, 0
	ldi	mux, 0x01
	clr	button
	clr	flags
	ldi	count, 1 ;avoid delay after first wake-up
	
	;go to standby
	sbic	PINB, 3 ;always-on mode?
	rjmp	power_off
	rjmp	power_on
	
;-------------------------------------------------------------------------------
	
ani_start:
	ldi	ZL, LOW(animations*2)
	ldi	ZH, HIGH(animations*2)
ani_loop:
	lpm	temp, Z+
	cpi	temp, 0x80 ;1000.0000 - rotate
	breq	ani_rotate
	brlo	ani_fill ;0xxx.xxxx - fill LEDs with value (or end)
	cpi	temp, 0x81 ;1000.0001 - fade
	breq	ani_fade
	cpi	temp, 0x82 ;1000.0010 - delay
	breq	ani_delay
	cpi	temp, 0x83 ;1000.0011 - repeat, repeat end
	brne	PC+2
	rjmp	ani_repeat
	cpi	temp, 0x84 ;1000.0100 - end of animations
	breq	ani_start
	;1000.0100 ~ 1111.1111 reserved for future extensions
	rjmp	ani_loop
	
	
ani_delay:
	;delay: count * 10 ms (for count = 0: 256 * 10 ms)
	lpm	count, Z+
ani_delay_loop:
	rcall	delay_10ms
	sbrc	flags, fButtonShort
	rcall	sound
	sbrs	flags, fButtonLong
	rjmp	ani_delay_end
	sbic	PINB, 3 ;always-on mode?
	rjmp	power_off
power_off_return:
ani_delay_end:
	dec	count
	brne	ani_delay_loop
	rjmp	ani_loop
	
ani_fade:
	;fade
	lpm	temp, Z+
	ldi	YL, LOW(RAM_Frame1)
	ldi	YH, HIGH(RAM_Frame1)
	ldi	count, 34
ani_fade_loop:
	ld	temp2, Y
	add	temp2, temp
	brpl	PC+2
	ldi	temp2, 0
	cpi	temp2, 8
	brlo	PC+2
	ldi	temp2, 7
	st	Y+, temp2
	dec	count
	brne	ani_fade_loop
	rjmp	ani_loop
	
ani_fill:
	;0aaa.aaal, llll.lvvv - fill l LEDs, starting at address a, with value v
	lpm	count, Z+
	ldi	temp2, 0
	lsr	temp
	ror	count
	ror	temp2
	lsr	count
	ror	temp2
	lsr	count
	ror	temp2
	lsr	temp2
	swap	temp2
	mov	YL, temp
	ldi	YH, 0
	subi	YL, LOW(-RAM_Frame1)
	sbci	YH, HIGH(-RAM_Frame1)
ani_fill_loop:
	cpi	YL, LOW(RAM_Frame1+34) ;wrap around for address
	brlo	PC+2
	sbiw	YH:YL, 34
	st	Y+, temp2 ;set LED
	dec	count
	brne	ani_fill_loop
	rjmp	ani_loop
	
ani_rotate:
	;rotate
	;copy LED values to backup framebuffer at shifted address
	lpm	temp, Z+
	ldi	XL, LOW(RAM_Frame1)
	ldi	XH, HIGH(RAM_Frame1)
	mov	YL, temp
	ldi	YH, 0
	subi	YL, LOW(-RAM_Frame2)
	sbci	YH, HIGH(-RAM_Frame2)
	ldi	count, 34
ani_rotate_loop1:
	ld	temp, X+
	st	Y+, temp
	cpi	YL, LOW(RAM_Frame2+34)
	brne	PC+2
	sbiw	YH:YL, 34
	dec	count
	brne	ani_rotate_loop1
	;copy shifted values back to displayed framebuffer
	ldi	YL, LOW(RAM_Frame1)
	ldi	YH, HIGH(RAM_Frame1)
	ldi	count, 34
ani_rotate_loop2:
	ldd	temp, Y+(RAM_Frame2-RAM_Frame1)
	st	Y+, temp
	dec	count
	brne	ani_rotate_loop2
	rjmp	ani_loop
	
ani_repeat:
	;repeat
	lpm	temp, Z+
	cpi	temp, 0
	breq	ani_repeat_end
	push	repeat ;store repeat counter for possible surrounding loop
	push	ZL ;store start-of-repeat-block address
	push	ZH
	mov	repeat, temp
	rjmp	ani_loop
	
ani_repeat_end:
	;repeat end
	dec	repeat
	breq	ani_repeat_complete
	pop	ZH ;read start-of-repeat-block address from stack
	pop	ZL
	push	ZL
	push	ZH
	rjmp	ani_loop
ani_repeat_complete:
	pop	temp ;remove start-of-repeat-block address from stack
	pop	temp
	pop	repeat ;restore repeat counter for possible surrounding loop
	rjmp	ani_loop
	
;--------------------

power_off:
	;disable PWM/mux timer
	ldi	temp, 0 ;stop timer
	out	TCCR0, temp
	ldi	temp, 0 ;disable timer overflow interrupt
	out	TIMSK, temp
	
	;LEDs off
	ldi	temp, 0x00
	out	PORTC, temp
	ldi	temp, 0x80
	out	PORTD, temp
	
	;backup stackpointer
	in	spl_backup, SPL
	in	sph_backup, SPH
	
	ori	flags, 1<<fExpectWdr
power_off_sleep1:
	ldi	wakeup_count, 4
power_off_sleep2:
	;sleep
	ldi	temp, 1<<SE | 1<<SM1
	out	MCUCR, temp
	sleep
wdr_wakeup:
	
	;check if button is being pressed
	sbic	PINB, 0
	rjmp	power_off_sleep1
	dec	wakeup_count
	brne	power_off_sleep2
	
	;disable sleeping
	ldi	temp, 0
	out	MCUCR, temp
	andi	flags, ~(1<<fExpectWdr)
	
	;restore stackpointer
	out	SPL, spl_backup
	out	SPH, sph_backup
	
power_on:
	;re-enable PWM/mux timer
	ldi	temp, 0x03 ;Clk/64
	out	TCCR0, temp
	ldi	temp, 1<<TOIE0
	out	TIMSK, temp
	sei
	
	ldi	button, BUTTON_LONG_TIME + 1
	andi	flags, ~(1<<fButtonShort | 1<<fButtonLong)
	
	rjmp	power_off_return

;--------------------

sound:

	;backup registers
	push	count
	push	ZL
	push	ZH
	
	;enable PWM on OC1A/OC1B, fast PWM mode, non-inv, 8 bit
	ldi	temp, 0
	out	TCNT1H, temp
	out	TCNT1L, temp
	ldi	temp, 0
	out	OCR1AH, temp
	ldi	temp, 0x80 ;idle: 50% duty cycle
	out	OCR1AL, temp
	ldi	temp, 0
	out	OCR1BH, temp
	ldi	temp, 0x80 ;idle: 50% duty cycle
	out	OCR1BL, temp
	ldi	temp, 1<<WGM10 | 1<<COM1A1 | 1<<COM1B1 ;fast PWM, non-inv.
	out	TCCR1A, temp
	ldi	temp, 1<<WGM12 | 0x01 ;fast PWM, Clk/1
	out	TCCR1B, temp
	
	;enable amplifier (prototype)
	cbi	PORTD, 7
	
	;backup framebuffer
	ldi	ZH, HIGH(RAM_Frame1+34)
	ldi	ZL, LOW(RAM_Frame1+34)
	ldi	count, 34
sound_leds_backup:
	ld	temp, -Z
	std	Z+(RAM_Frame2-RAM_Frame1), temp 
	dec	count
	brne	sound_leds_backup
	
	;fade all LEDs on
fade_on:
	ldi	count, 1
fade_on_loop1:
	ldi	ZH, HIGH(RAM_Frame1)
	ldi	ZL, LOW(RAM_Frame1)
	ldi	temp2, 34
fade_on_loop2:
	ld	temp, Z
	cp	temp, count
	brsh	PC+2
	st	Z, count
	adiw	ZH:ZL, 1
	dec	temp2
	brne	fade_on_loop2
	rcall	delay_25ms
	inc	count
	cpi	count, 8
	brlo	fade_on_loop1
	
	;delay until amplifier ready (prototype)
	ldi	temp, 50
	rcall	delay_adj
	
	;start playback
	ldi	ZH, HIGH(audio*2)
	ldi	ZL, LOW(audio*2)
	ldi	temp, 1<<TOV1 ;clear timer overflow interrupt flag
	out	TIFR, temp
	in	temp, TIMSK
	ori	temp, 1<<TOIE1 ;enable timer overflow interrupt
	out	TIMSK, temp
	
	;wait until playback finished
sound_wait:
	in	temp, TIMSK
	sbrc	temp, TOIE1
	rjmp	sound_wait
	
	;disable amplifier (prototype)
	sbi	PORTD, 7
	
	;delay until amplifier off (avoid popping noise) (prototype)
	ldi	temp, 50
	rcall	delay_adj
	
fade_off:
	ldi	count, 6
fade_off_loop1:
	ldi	ZH, HIGH(RAM_Frame1)
	ldi	ZL, LOW(RAM_Frame1)
	ldi	temp2, 34
fade_off_loop2:
	ldd	temp, Z+(RAM_Frame2-RAM_Frame1)
	cp	count, temp
	brlo	PC+2
	st	Z, count
	adiw	ZH:ZL, 1
	dec	temp2
	brne	fade_off_loop2
	rcall	delay_25ms
	subi	count, 1
	brcc	fade_off_loop1
	
;	;restore previous frame
;	ldi	ZH, HIGH(RAM_Frame1)
;	ldi	ZL, LOW(RAM_Frame1)
;	ldi	count, 34
;sound_leds_off:
;	ldd	temp, Z+(RAM_Frame2-RAM_Frame1)
;	st	Z+, temp
;	dec	count
;	brne	sound_leds_off
	
	;disable PWM
	ldi	temp, 0
	out	TCCR1B, temp
	out	TCCR1A, temp
	
	;restore registers
	pop	ZH
	pop	ZL
	pop	count

	andi	flags, ~(1<<fButtonShort)
	ret
	
;--------------------

delay_10ms:
	;approx. 10 ms delay
	ldi	temp, 104
	rjmp	delay_adj
delay_25ms:
	;approx. 25 ms delay
	clr	temp
delay_adj:
	;adjustable delay: ~100 us * temp
	clr	temp2
delay_loop:
	dec	temp2
	brne	delay_loop
	dec	temp
	brne	delay_loop
	ret
	